from apex_client.objects import ApexObject
from apex_client.objects.events.mixins import EventParticipantMixin


class SessionParticipant(EventParticipantMixin, ApexObject):
    valid_properties = ['id', 'participant_name', 'participant_email', 'participant_cell_phone_number',
                        'participant_phone_number', 'event', 'event_name', 'session', 'session_name', 'company_name',
                        'group', 'group_name', 'status', 'sessions', 'hours', 'event_days', 'invoice', 'portal',
                        'portal_name', 'event_identifier', 'attendance', 'coach', 'coach_name', 'coach_email',
                        'hours_daily', 'invoice_status', 'invoice_total_price', 'invoice_total_price_default_currency']
    date_properties = ['event_start_date', 'event_end_date', 'session_start_date', 'session_end_date', 'booked_date',
                       'confirmation_sent', 'invitation_sent']

    def __str__(self):
        return f"{self.event_name}: {self.participant_name} - {self.session_name} - {self.status}"



from apex_client.objects import ApexObject
from apex_client.objects.events import Event

from .portal import Portal


class PortalEvent(ApexObject):
    valid_properties = ['id', 'uuid', 'portal_id', 'event_id', 'portal_name', 'is_available', 'max_participants',
                        'booked_participants_count', 'identifier', 'event_name', 'is_active', 'number_of_days']

    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.event_object = None
        self.portal_object = None

    @property
    def event(self):
        if not self.event_object and self.event_id:
            self.event_object = Event.get(self.event_id, True)

            if not self.event_object:
                self.event_object = Event.get(self.event_id)

        return self.event_object

    @property
    def portal(self):
        if not self.portal_object and self.portal_id:
            self.portal_object = Portal.get(self.portal_id)

        return self.portal_object

    def __str__(self):
        return f"{self.event_name}: {self.portal_name}"



from apex_client.objects import ApexObject


class MailTemplate(ApexObject):
    valid_properties = ['id', 'uuid', 'template_type', 'content', 'subject', 'sender_name', 'sender_email']
    mapped_properties = {
        'event': 'event_id'
    }

    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.event_object = None

    @property
    def event(self):
        from .event import Event

        if not self.event_object and self.event_id:
            self.event_object = Event.get(self.event_id)

        return self.event_object



from apex_client.objects import ApexObject


class EventArea(ApexObject):
    valid_properties = ['id', 'uuid', 'area_id', 'event_id', 'area_name', 'area_slug', 'level']

    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.event_object = None
        self.area_object = None

    @property
    def event(self):
        from .event import Event

        if self.event_object is None and self.event_id:
            self.event_object = Event.get(self.event_id)

        return self.event_object

    @property
    def area(self):
        from .area import Area

        if self.area_object is None and self.area_id:
            self.area_object = Area.get(self.area_id)

        return self.area_object

    def __str__(self):
        return f"{self.area_name}: {self.event_id}"



from typing import Optional

from apex_client.objects import ApexObject
from apex_client.request import ApexRequest

from .day import Day
from .session_participant import SessionParticipant


class Session(ApexObject):
    valid_properties = ['id', 'identifier', 'event_id', 'order', 'name', 'uuid', 'parent', 'is_active', 'is_template',
                        'number_of_days', 'timezone', 'venue', 'venue_name', 'venue_city', 'min_participants',
                        'max_participants', 'language']
    date_properties = ['date_created', 'date_modified', 'start_date', 'end_date']

    def __init__(self, json_data):
        super().__init__(json_data)

        self.days = []
        self.session_participants = []

        if 'days' in json_data.keys():
            for item in json_data['days']:
                self.days.append(Day(item))

    def get_participants(self, status: Optional[str] = None) -> list:
        if not self.is_template and (not self.session_participants or status):
            self.session_participants = []
            results = ApexRequest.search_all(f'events/{self.parent}/sessions/{self.id}/participants/',
                                             {'status': status} if status else None)

            if results:
                for item in results:
                    self.session_participants.append(SessionParticipant(item))

        return self.session_participants

    def __str__(self):
        return f"{self.event_id}#{self.order} - {self.name}"



from apex_client.objects import ApexObject


class Day(ApexObject):
    valid_properties = ['id', 'uuid', 'name', 'event', 'booked_participant_count', 'preliminary_participant_count',
                        'reserved_participant_count']
    date_properties = ['start_date', 'end_date']

    def __init__(self, json_data):
        super().__init__(json_data)

        self.session_object = None

    @property
    def session(self):
        from .session import Session

        if not self.session_object and self.event:
            self.session_object = Session.get(self.event)

        return self.session_object

    def __str__(self):
        return f"{self.start_date.isoformat() if self.start_date else ''} - " \
               f"{self.end_date.isoformat() if self.end_date else ''}"



from typing import Optional

from apex_client.objects import ApexObject
from .currency import Currency


class Price(ApexObject):
    valid_properties = ['id', 'uuid', 'price', 'currency_name']
    mapped_properties = {
        'currency': 'currency_id'
    }

    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.currency_object = None

    @classmethod
    def get(cls, price_id: str) -> Optional['Price']:
        return Price._get(f'invoices/prices/{price_id}/')

    @property
    def currency(self):
        if not self.currency_object and self.currency_id:
            self.currency_object = Currency.get(self.currency_id)

        return self.currency_object

    def __str__(self):
        return f"{self.price} {self.currency_name}"



from apex_client.objects import ApexObject


class Address(ApexObject):
    valid_properties = ['id', 'uuid', 'type', 'position', 'extra_address_field', 'address_row_1', 'address_row_2',
                        'address_row_3', 'address_row_4', 'zip_code', 'city', 'state', 'country']
    date_properties = ['date_created', 'date_modified']

    def __str__(self):
        return f"{self.address_row_1 if self.address_row_1 else ''} {self.city}"



from typing import Optional

from apex_client.objects import ApexObject
from apex_client.objects.events import Event
from apex_client.objects.participants import Company, Participant
from apex_client.request import ApexRequest
from .currency import Currency
from .price import Price


class Invoice(ApexObject):
    valid_properties = ['id', 'uuid', 'event_name', 'company_name', 'participant_name', 'status', 'status_display',
                        'payment_type', 'currency_id', 'currency_shortname', 'our_reference', 'our_reference_name',
                        'our_reference_email', 'attention', 'attention_email', 'reference', 'event_economy_account_1',
                        'event_acconomy_account_2', 'financial_id', 'total_cost', 'total_cost_default_currency']
    date_properties = ['date', ]
    mapped_properties = {
        'company': 'company_id',
        'event': 'event_id',
        'participant': 'participant_id'
    }

    def __init__(self, json_data: dict):
        from .invoice_row import InvoiceRow

        super().__init__(json_data)

        self.company_object = None
        self.currency_object = None
        self.event_object = None
        self.invoice_row_objects = []
        self.participant_object = None

        if 'invoice_rows' in json_data.keys():
            for item in json_data['invoice_rows']:
                if isinstance(item, dict):
                    self.invoice_row_objects.append(InvoiceRow(item))

        if 'total_price' in json_data.keys():
            self.total_price = Price(json_data['total_price'])
        else:
            self.total_price = Price({})

        if 'total_price_default_currency' in json_data.keys():
            self.total_price_default_currency = Price(json_data['total_price_default_currency'])
        else:
            self.total_price_default_currency = Price({})

    @classmethod
    def get(cls, invoice_id: str) -> Optional['Invoice']:
        return Invoice._get(f'invoices/{invoice_id}/')

    @classmethod
    def search(cls, params: dict) -> list:
        return ApexRequest.search_all('invoices/', params)

    @property
    def company(self) -> Optional['Company']:
        if not self.company_object and self.company_id:
            self.company_object = Company.get(self.company_id)

        return self.company_object

    @property
    def event(self) -> Optional['Event']:
        if not self.event_object and self.event_id:
            self.event_object = Event.get(self.event_id)

        return self.event_object

    @property
    def participant(self) -> Optional['Participant']:
        if not self.participant_object and self.participant_id:
            self.participant_object = Participant.get(self.participant_id)

        return self.participant_object

    @property
    def invoice_rows(self) -> list:
        from .invoice_row import InvoiceRow

        if not self.invoice_row_objects:
            results = ApexRequest.search_all(f'invoices/{self.id}/rows/')

            if results:
                for item in results:
                    self.invoice_row_objects.append(InvoiceRow(item))

        return self.invoice_row_objects

    @property
    def currency(self) -> Optional['Currency']:
        if not self.currency_object and self.currency_id:
            self.currency_object = Currency.get(self.currency_id)

        return self.currency_object

    def __str__(self):
        return f"{self.event_name}: {self.participant_name}"



from typing import Optional

from apex_client.request import ApexRequest

from apex_client.objects import ApexObject
from apex_client.objects.events.mixins import EventParticipantMixin


class EventParticipant(EventParticipantMixin, ApexObject):
    STATUSES = ['booked', 'canceled', 'preliminary', 'reserve', 'rescheduled']

    valid_properties = ['id', 'participant_name', 'participant_email', 'participant_cell_phone_number',
                        'participant_phone_number', 'event', 'event_name', 'company_name', 'group',
                        'group_name', 'status', 'sessions', 'hours', 'event_days', 'invoice', 'portal', 'portal_name',
                        'event_identifier', 'attendance', 'coach', 'coach_name', 'coach_email', 'hours_daily',
                        'invoice_status', 'invoice_total_price', 'invoice_total_price_default_currency']
    date_properties = ['event_start_date', 'event_end_date', 'booked_date', 'confirmation_sent', 'invitation_sent']

    @classmethod
    def get(cls, event_id: int, participant_id: int) -> Optional['EventParticipant']:
        response = ApexRequest.get(f'events/{event_id}/participants/{participant_id}/')

        if response['status_code'] == 200:
            return cls(response['response'])

        return None

    def update(self) -> bool:
        if self.status in EventParticipant.STATUSES:
            response = ApexRequest.patch(f'events/{self.event}/participants/{self.participant}/',
                                         {'status': self.status, 'hours_daily': self.hours_daily})

            if response['status_code'] == 200:
                return True

        return False

    def __str__(self):
        return f"{self.event_name}: {self.participant_name} - {self.status}"



from typing import Optional

from apex_client.objects import ApexObject

from .company import Company


class Participant(ApexObject):
    valid_properties = ['id', 'uuid', 'first_name', 'last_name', 'full_name', 'email', 'is_active', 'company_name',
                        'gender']
    date_properties = ['date_created', 'date_modified', 'deletion_date']

    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.company = None

        if 'company' in json_data and json_data['company']:
            self.company = Company(json_data['company'])

    @classmethod
    def get(cls, participant_id: str) -> Optional['Participant']:
        return Participant._get(f'participants/{participant_id}/')

    def __str__(self):
        return f"{self.full_name} ({self.email})"



from apex_client.objects.participants import Company, Participant


class EventParticipantMixin:
    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.participant_id = json_data['participant']
        self.company_id = json_data['company']

        self.participant_object = None
        self.company_object = None

    @property
    def participant(self):
        if not self.participant_object and self.participant_id:
            self.participant_object = Participant.get(self.participant_id)

        return self.participant_object

    @property
    def company(self):
        if not self.company_object and self.company_id:
            self.company_object = Company.get(self.company_id)

        return self.company_object



import functools
import logging
import urllib.parse
from typing import Optional

from django.conf import settings

from apex_auth.apex_request import ApexRequest as AuthRequest
from requests import Response, get, patch

LOGGER = logging.getLogger(__name__)


def return_json(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs) -> dict:
        response = f(self, *args, **kwargs)

        if response.status_code >= 299:
            if response.status_code >= 500:
                LOGGER.error(f"Error received from Apex server: {response.status_code} Server error")
                return {
                    'status_code': response.status_code,
                    'response': {
                        'error': 'Server error'
                    }
                }
            else:
                LOGGER.error(f"Error received from Apex server: {response.status_code}: {response.json()}")

        return {
            'status_code': response.status_code,
            'response': response.json()
        }

    return wrapper


class ApexRequest:
    base_url = f"{settings.APEX_SERVER}api/v1/"

    @staticmethod
    def _get_headers(data=None):
        headers = AuthRequest.create_request_headers(settings.APEX_PUBLIC_KEY, settings.APEX_PRIVATE_KEY, data)

        headers.update({
            "Content-type": "application/json"
        })

        return headers

    @staticmethod
    @return_json
    def get(path: str, params: Optional[dict] = None) -> Response:
        LOGGER.info(f'GET: {path}, params: {params}')
        path = urllib.parse.urljoin(ApexRequest.base_url, path, allow_fragments=True)

        return get(path, params, headers=ApexRequest._get_headers())

    @staticmethod
    @return_json
    def patch(path: str, value: dict) -> Response:
        LOGGER.info(f'PATCH: {path}, value: {value}')
        path = urllib.parse.urljoin(ApexRequest.base_url, path, allow_fragments=True)

        return patch(path, value, headers=ApexRequest._get_headers(value))

    @staticmethod
    def search(path: str, params: dict) -> dict:
        LOGGER.info(f'search: {path}, params: {params}')

        if 'limit' not in params:
            params.update({'limit': 25})

        if 'offset' not in params and params.get('limit') != 'null':
            params.update({'offset': 0})

        return ApexRequest.get(path, params)

    @staticmethod
    def search_all(path: str, params: Optional[dict] = None) -> Optional[list]:
        LOGGER.info(f'search all: {path}, params: {params}')
        done = False
        results = []

        if not params:
            params = {
                'limit': 25,
                'offset': 0
            }
        elif 'limit' not in params.keys():
            params.update({
                'limit': 25,
                'offset': 0
            })

        while not done:
            response = ApexRequest.search(path, params)

            if response['status_code'] == 200:
                if response['response']['next'] is None:
                    done = True
                else:
                    params['offset'] += params['limit']

                results.extend(response['response']['results'])
            else:
                return None

        return results



from apex_client.objects import ApexObject


class Currency(ApexObject):
    valid_properties = ['id', 'uuid', 'name', 'short_name']

    @classmethod
    def get(cls, currency_id: str):
        return Currency._get(f'invoices/currencies/{currency_id}/')

    def __str__(self):
        return self.name



from typing import Optional

from apex_client.objects import ApexObject


class Network(ApexObject):
    valid_properties = ['id', 'name']

    @classmethod
    def get(cls, network_id: str) -> Optional['Network']:
        return Network._get(f'networks/{network_id}/')

    def __str__(self):
        return self.name



from typing import Optional

from apex_client.objects import ApexObject


class Type(ApexObject):
    valid_properties = ['id', 'uuid', 'name', 'number_of_events']
    date_properties = ['date_created', 'date_modified']

    @classmethod
    def get(cls, type_id: str) -> Optional['Type']:
        return Type._get(f"events/types/{type_id}/")

    def __str__(self):
        return f"{self.name}"



from apex_client.objects import ApexObject

from .price import Price


class InvoiceRow(ApexObject):
    valid_properties = ['id', 'uuid', 'economy_account_1', 'economy_account_2', 'description', 'discount']
    mapped_properties = {
        'invoice': 'invoice_id'
    }

    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.price_object = None
        self.price_id = None

        self.invoice_object = None
        self.invoice_id = None

        if 'price' in json_data.keys():
            if isinstance(json_data['price'], dict):
                self.price_object = Price(json_data['price'])
            else:
                self.price_id = json_data['price']

    @classmethod
    def get(cls, invoice_id: str, invoice_row_id: str):
        return InvoiceRow._get(f'invoices/{invoice_id}/rows/{invoice_row_id}/')

    @property
    def price(self):
        if not self.price_object and self.price_id:
            self.price_object = Price.get(self.price_id)

        return self.price_object

    @property
    def invoice(self):
        from .invoice import Invoice

        if not self.invoice_object and self.invoice_id:
            self.invoice_object = Invoice.get(self.invoice_id)

        return self.invoice_object

    def __str__(self):
        return f"{self.id}: {self.description}"



from .area import Area
from .day import Day
from .event import Event
from .event_area import EventArea
from .event_participant import EventParticipant
from .mail_template import MailTemplate
from .session import Session
from .session_participant import SessionParticipant
from .type import Type



from typing import Optional

from apex_client.objects import ApexObject
from apex_client.objects.participants import Company
from apex_client.request import ApexRequest


class Portal(ApexObject):
    valid_properties = ['id', 'uuid', 'name', 'url', 'type', 'styles', 'logo', 'bookable', 'company_name', 'events']
    date_properties = ['date_created', 'date_modified']

    def __init__(self, json_data: dict):
        super().__init__(json_data)
        self.areas = []
        self.portal_events = []

        if 'company' in json_data.keys():
            if isinstance(json_data['company'], dict):
                self.company_object = Company(json_data['company'])
                self.company_id = self.company_object.id
            else:
                self.company_object = None
                self.company_id = json_data['company']

    @classmethod
    def get(cls, portal_id: str) -> Optional['Portal']:
        return Portal._get(f'portals/{portal_id}/')

    @classmethod
    def search(cls, name: str = None) -> Optional[list]:
        params = None

        if name:
            params.update({
                "name": name
            })

        return ApexRequest.search_all('portals/', params)

    @property
    def company(self):
        if not self.company_object and self.company_id:
            self.company_object = Company.get(self.company_id)

        return self.company_object

    def __str__(self):
        return self.name

    def get_portal_events(self, reload: bool = False) -> list:
        from .portal_event import PortalEvent

        if not self.portal_events or reload:
            self.portal_events = []
            portal_events = ApexRequest.search_all(f'portals/{self.id}/events/')

            if portal_events is not None:
                for portal_event in portal_events:
                    self.portal_events.append(PortalEvent(portal_event))

        return self.portal_events

    def get_areas(self, reload: bool = False) -> list:
        from apex_client.objects.events import Area

        if not self.areas or reload:
            self.areas = []

            areas = ApexRequest.search_all(f'websites/{self.id}/areas/', {'limit': 'null'})

            if areas is not None:
                for area in areas:
                    self.areas.append(Area(area))

        return self.areas



import datetime
from typing import Optional

from django.template.defaultfilters import date as _date

from apex_client.objects import ApexObject
from apex_client.objects.exceptions import MultipleObjectsReturned
from apex_client.request import ApexRequest

from .event_area import EventArea
from .event_participant import EventParticipant
from .mail_template import MailTemplate
from .session import Session
from .type import Type


class Event(ApexObject):
    valid_properties = ['id', 'slug', 'identifier', 'event_id', 'order', 'name', 'uuid', 'is_active', 'is_template',
                        'description', 'parent', 'number_of_days', 'timezone', 'venue', 'venue_name', 'venue_city',
                        'min_participants', 'max_participants', 'language', 'invoiced_date', 'promote_program',
                        'portal_participants', 'booked_participant_count']
    date_properties = ['date_created', 'date_modified', 'start_date', 'end_date']

    def __init__(self, json_data: dict):
        from apex_client.objects.portals import PortalEvent

        super().__init__(json_data)

        keys = json_data.keys()

        self.areas = []
        self.mail_templates = []
        self.portals = []
        self.sessions = []
        self.types = []

        if 'areas' in keys:
            for item in json_data['areas']:
                if isinstance(item, dict):
                    self.areas.append(EventArea(item))

        if 'mail_templates' in keys:
            for item in json_data['mail_templates']:
                if isinstance(item, dict):
                    self.mail_templates.append(MailTemplate(item))

        if 'portals' in keys:
            for item in json_data['portals']:
                if isinstance(item, dict):
                    self.portals.append(PortalEvent(item))

        if 'sessions' in keys:
            for item in json_data['sessions']:
                if isinstance(item, dict):
                    self.sessions.append(Session(item))

        if 'types' in keys:
            for item in json_data['types']:
                if isinstance(item, dict):
                    self.types.append(Type(item))

    @classmethod
    def get(cls, event_id: str, template: Optional[bool] = False) -> Optional['Event']:
        if template:
            return Event._get(f'templates/{event_id}/')
        else:
            return Event._get(f'events/{event_id}/')

    @classmethod
    def get_by_slug(cls, slug: str) -> Optional['Event']:
        results = ApexRequest.search_all(f'templates/', {
            'slug': slug,
            'is_active': True,
            'extra_fields': 'sessions,prices,areas,mail_templates,portals,types'
        })

        if results is None or len(results) == 0:
            return None

        if len(results) == 1:
            return cls(results[0])
        else:
            raise MultipleObjectsReturned

    @property
    def dates(self):
        dates = []

        for session in self.sessions:
            if len(session.days) == 1:
                dates.append(f"{_date(session.start_date, 'j M')}")
            else:
                if session.start_date.month == session.end_date.month:
                    dates.append(f"{_date(session.start_date, 'j')}-{_date(session.end_date, 'j M')}")
                else:
                    dates.append(f"{_date(session.start_date, 'j M')}-{_date(session.end_date, 'j M')}")

        return ", ".join(dates)

    def get_events(self, coming: bool = False, days_before: int = 1) -> list:
        events = []

        if self.is_template:
            params = {
                'ordering': 'start_date',
                'extra_fields': 'sessions,prices,portals'
            }

            if coming:
                now = datetime.datetime.utcnow() + datetime.timedelta(days=days_before)
                params.update({'start_date': now.strftime("%Y-%m-%d")})

            results = ApexRequest.search_all(f'templates/{self.id}/events/', params)

            for item in results:
                events.append(Event(item))

        return events

    def get_participants(self, status=None) -> list:
        participants = []

        if not self.is_template:
            params = {}

            if status:
                params.update({'status': status})

            results = ApexRequest.search_all(f'events/{self.id}/participants/', params)

            for item in results:
                participants.append(EventParticipant(item))

        return participants

    def is_bookable(self, portal_id: str = None) -> bool:
        bookable = False

        if self.booked_participant_count < self.max_participants:
            bookable = True

        if portal_id and self.portal_participants and str(portal_id) in self.portal_participants.keys():
            booked_participants = self.portal_participants[str(portal_id)]
            max_participants = 0

            for portal in self.portals:
                if portal.portal_id == portal_id:
                    max_participants = portal.max_participants

                if portal.event_id == self.id and portal.portal_id == portal_id:
                    max_participants = portal.max_participants
                    break

            if booked_participants < max_participants:
                bookable = True
            else:
                bookable = False

        return bookable

    def has_few_places_left(self, portal_id: str = None) -> bool:
        few_places_left = False

        if self.booked_participant_count > self.max_participants - 3:
            few_places_left = True

        if not few_places_left and portal_id and self.portal_participants and \
                str(portal_id) in self.portal_participants.keys():
            booked_participants = self.portal_participants[str(portal_id)]
            max_participants = 0

            for portal in self.portals:
                if portal.portal_id == portal_id:
                    max_participants = portal.max_participants

                if portal.event_id == self.id and portal.portal_id == portal_id:
                    max_participants = portal.max_participants
                    break

            if booked_participants > max_participants - 3:
                few_places_left = True

        return few_places_left

    def __str__(self):
        return f"{self.event_id}: {self.name}"



from typing import Optional

from apex_client.objects import ApexObject

from .address import Address
from .network import Network


class Company(ApexObject):
    valid_properties = ['id', 'uuid', 'name', 'is_active', 'website', 'corporate_id', 'vat_number', 'default_currency',
                        'allow_invoicing', 'terms_of_payment', 'group_invoices', 'invoice_days_before_event',
                        'number_of_employees', 'city', 'default_currency_display', 'status', 'invoice_company',
                        'invoice_company_name', 'booked_events']
    date_properties = ['date_created', 'date_modified']

    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.address_ids = []
        self.address_objects = []

        self.network_ids = []
        self.network_objects = []

        if 'addresses' in json_data.keys():
            for item in json_data['addresses']:
                if isinstance(item, dict):
                    self.address_objects.append(Address(item))
                else:
                    self.address_ids.append(item)

        if 'networks' in json_data.keys():
            for item in json_data['networks']:
                if isinstance(item, dict):
                    self.network_objects.append(Network(item))
                else:
                    self.network_ids.append(item)

    @classmethod
    def get(cls, company_id: str) -> Optional['Company']:
        return Company._get(f'companies/{company_id}/')

    @property
    def addresses(self):
        if not len(self.address_objects) and len(self.address_ids):
            for address_id in self.address_ids:
                address = self.get_address(address_id)
                if address:
                    self.address_objects.append(address)

        return self.address_objects

    def get_address(self, address_id: str) -> Optional[Address]:
        return Address._get(f'companies/{self.id}/addresses/{address_id}/')

    @property
    def networks(self):
        if not len(self.network_objects) and len(self.network_ids):
            for network_id in self.network_ids:
                network = Network.get(network_id)
                if network:
                    self.network_objects.append(network)

        return self.network_objects

    def __str__(self):
        return f"{self.uuid}: {self.name}"



from typing import Optional

from dateutil import parser

from apex_client.request import ApexRequest


class ApexObject:
    valid_properties = []
    date_properties = []
    mapped_properties = {}

    def __init__(self, json_data: dict):
        self.dynamic_properties = []

        for key in self.valid_properties:
            setattr(self, key, json_data.get(key))

        for key in self.mapped_properties.keys():
            setattr(self, self.mapped_properties[key], json_data.get(key))

        if 'dynamic_fields' in json_data:
            for dynamic_field in json_data.get('dynamic_fields'):
                dynamic_dict = {
                    'network_project_id': dynamic_field.get('network_project_id'),
                    'network_project_name': dynamic_field.get('network_project_name'),
                    'fields': {}
                }

                for field in dynamic_field['fields']:
                    dynamic_dict['fields'].update({
                        field['property']: {
                            'network_id': field['network_id'],
                            'type': field['type'],
                            'value': field['value']
                        }
                    })

                self.dynamic_properties.append(dynamic_dict)

        for key in self.date_properties:
            if json_data.get(key):
                setattr(self, key, parser.parse(json_data.get(key)))
            else:
                setattr(self, key, None)

    @classmethod
    def _get(cls, path: str) -> Optional['ApexObject']:
        response = ApexRequest.get(path)

        if response['status_code'] == 200:
            return cls(response['response'])

        return None

    def _get_dynamic_property_from_dict(self, index, key, default_value=None):
        try:
            dynamic_property = self.dynamic_properties[index]['fields'].get(key)

            if dynamic_property:
                value = dynamic_property.get('value')

                if value is not None:
                    return value

        except IndexError:
            pass

        if default_value is not None:
            return default_value

        return None

    def get_dynamic_property(self, key, default_value=None):
        for item in self.dynamic_properties[::-1]:
            if key in item['fields'].keys():
                return self._get_dynamic_property_from_dict(self.dynamic_properties.index(item), key, default_value)

        return self._get_dynamic_property_from_dict(0, key, default_value)

    def has_dynamic_property(self, key) -> bool:
        for item in self.dynamic_properties:
            if key in item['fields'].keys():
                return True

        return False

    def get_dynamic_property_display(self, key):
        dynamic_property = None

        for item in self.dynamic_properties:
            if key in item['fields'].keys():
                dynamic_property = item['fields'][key]
                break

        if dynamic_property:
            value = self.get_dynamic_property(key)

            if value and dynamic_property.get('options') and dynamic_property.get('options').get('choices'):
                for choice in dynamic_property['options']['choices']:
                    if str(choice[0]) == str(value):
                        return choice[1]

        return None

    def get_dynamic_fields(self):
        dynamic_fields = []

        for item in self.dynamic_properties:
            dynamic_property = {
                'network_project_id': item['network_project_id'],
                'network_project_name': item['network_project_name'],
                'fields': [
                    {
                        'network_id': field['network_id'],
                        'property': key,
                        'value': field['value']
                    } for key, field in item['fields'].items()
                ]
            }

            dynamic_fields.append(dynamic_property)

        return dynamic_fields

    def __eq__(self, other):
        return type(self) == type(other) and self.id == other.id

    def __repr__(self):
        return f"<{self.__class__.__name__}: {str(self)}>"

    def __str__(self):
        return "Empty class"

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)

        except AttributeError as e:
            if name in self.date_properties or name in self.valid_properties:
                return None
            else:
                raise e



from typing import Optional

from apex_client.objects import ApexObject
from .event import Event


class Area(ApexObject):
    valid_properties = ['id', 'uuid', 'name', 'slug', 'number_of_events', 'number_of_templates']
    date_properties = ['date_created', 'date_modified']

    def __init__(self, json_data: dict):
        super().__init__(json_data)

        self.area_templates = []
        self.area_events = []

        if 'area_templates' in json_data.keys():
            for item in json_data['area_templates']:
                self.area_templates.append(Event(item))

        if 'area_events' in json_data.keys():
            for item in json_data['area_events']:
                self.area_events.append(Event(item))

    @classmethod
    def get(cls, area_id: str) -> Optional['Area']:
        return Area._get(f'areas/{area_id}/')

    def __str__(self):
        return self.name
